# Converted to Python 3.5.2 on 01-25-17
# by: Jeremy Webster


# coding: utf-8
"""
Core columns and abstract classes representing the KB Core seismic schema.

ORM classes that inherit from these classes are guaranteed primary and unique
keys, column info dictionaries, and schema-qualified __tablename__ .

Examples
--------
Define a site class that points to a 'TA_site' table.

>>> import pisces.schema.kbcore as kb
>>>
>>> class Site(kb.Site):
>>>     __tablename__ = 'TA_site'

Define a site class that points to a 'jkmacc.my_site' table.

>>> import pisces.schema.kbcore as kb
>>>
>>> class Site(kb.Site):
>>>     __tablename__ = 'jkmacc.my_site'

Create this table, if it doesn't exist.

>>> import sqlalchemy as sa
>>> e = sa.create_engine('sqlite:///my.sqlite')
>>> if not Site.__table__.exists(e)
>>>     Site.__table__.create(e)

Define a new abstract class in this schema.

>>> import sqlalchemy as sa
>>> from sqlalchemy.ext.declarative import declared_attr
>>> import pisces.schema.kbcore as kb
>>>
>>> def strparse(s):
>>>     return s.strip()
>>>
>>> class Beam(kb.Base):
>>>     __abstract__ = True
>>>     @declared_attr
>>>     def __table_args__(cls):
>>>         return (sa.PrimaryKeyConstraint('wfid'),)
>>>
>>>     wfid = kb.wfid.copy()
>>>     azimuth = kb.azimuth.copy()
>>>     slo = sa.Column(Float(24),
>>>         info={'default': -1.0, 'parse': float, 'width': 7, 'format': '7.4f'})
>>>     filter = sa.Column(String(30),
>>>         info={'default': '-', 'parse': strparse, 'width': 30, 'format': '30.30s'})
>>>     recipe = sa.Column(String(15),
>>>         info={'default': '-', 'parse': strparse, 'width': 15, 'format': '15.15s'})
>>>     algorithm = kb.algorithm.copy()
>>>     auth = kb.auth.copy()
>>>     lddate = kb.lddate.copy()

"""
from datetime import datetime

from sqlalchemy import DateTime, Float, String, Integer
from sqlalchemy import Column
from sqlalchemy import PrimaryKeyConstraint, UniqueConstraint
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr

from pisces.schema.util import PiscesMeta
from pisces.schema.util import parse_str, parse_int, parse_float
from pisces.io.trace import wfdisc2trace

# TODO: check __table_args__ syntax for final empty {}
# TODO: use http://docs.sqlalchemy.org/en/rel_0_8/orm/extensions/hybrid.html?

Base = declarative_base(metaclass=PiscesMeta, constructor=None)

# COLUMN DEFINITIONS
# Generic SQLA types, compatible with different backends
# !! info dictionary defines the external representations (NumPy, text files)
#   and default values for the mapped class representation.
# XXX: for numeric types, maximum width is not enforced!
# NOTE: info['dtype'] for floats/ints should match the system default for Python
#       "{:f}".format(numpyfloat/int) to work, b/c it can use the builtin float
#       __format__().  use 'float' or 'int'
#       See:
#       http://stackoverflow.com/questions/16928644/floats-in-numpy-structured-array-and-native-string-formatting-with-format
# NOTE: column names are not defined here, they are assigned during declarative
#       table definitions.  i.e. amp.name is None until it is used later.  This
#       allows recycling columns as different names, like sta1 = sta.copy().

# TODO: write a parser that gets the width (and type?) from info['format']

# specialized string parsing functions

# TODO: use dateutil module to handle wildcard characters, etc.?
# https://docs.python.org/2.7/library/datetime.html#strftime-strptime-behavior
# https://docs.python.org/2/library/string.html#format-specification-mini-language
DATEFMT = '%Y-%m-%d %H:%M:%S'


def dtfn(s):
    try:
        val = datetime.strptime(s, DATEFMT)
    except ValueError:
        val = None
    return val


algorithm = Column(String(15),
                   info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

amp = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 11, 'format': '11.2f'})

ampid = Column(Integer, nullable=False,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

amptime = Column(Float(53), info={'default': -9999999999.999, 'parse': parse_float,
                                  'dtype': 'float', 'width': 17, 'format': '17.5f'})

amptype = Column(String(8),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

arid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

auth = Column(String(20),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a20', 'width': 20, 'format': '20.20s'})

azdef = Column(String(1),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

azimuth = Column(Float(24),
                 info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

azres = Column(Float(24),
               info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.1f'})

band = Column(String(1),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

belief = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 4, 'format': '4.2f'})

calib = Column(Float(24),
               info={'default': 1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

calper = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

calratio = Column(Float(24),
                  info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

chan = Column(String(8),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

chanid = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

clip = Column(String(1),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

commid = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

conf = Column(Float(24),
              info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 5, 'format': '5.3f'})

ctype = Column(String(4),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

datatype = Column(String(2),
                  info={'default': '-', 'parse': parse_str, 'dtype': 'a2', 'width': 2, 'format': '2.2s'})

deast = Column(Float(24),
               info={'default': 0.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

delaz = Column(Float(24),
               info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

delslo = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

delta = Column(Float(24),
               info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8.3f'})

deltaf = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.3f'})

deltim = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.3f'})

depdp = Column(Float(24),
               info={'default': -999, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

depth = Column(Float(24),
               info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

descrip = Column(String(50),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})

dfile = Column(String(32),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a32', 'width': 32, 'format': '32.32s'})

digital = Column(String(1),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

dir = Column(String(64),
             info={'default': '-', 'parse': parse_str, 'dtype': 'a64', 'width': 64, 'format': '64.64s'})

dnorth = Column(Float(24),
                info={'default': 0.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

dtype = Column(String(1),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

duration = Column(Float(24),
                  info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

edepth = Column(Float(24),
                info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

elev = Column(Float(24),
              info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

ema = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

emares = Column(Float(24),
                info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.1f'})

endtime = Column(Float(53), info={'default': 9999999999.999,
                                  'parse': parse_float, 'dtype': 'float', 'width': 17, 'format': '17.5f'})

esaz = Column(Float(24),
              info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

etype = Column(String(7),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a7', 'width': 7, 'format': '7.7s'})

evid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

evname = Column(String(32),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a32', 'width': 32, 'format': '32.32s'})

fm = Column(String(2),
            info={'default': '-', 'parse': parse_str, 'dtype': 'a2', 'width': 2, 'format': '2.2s'})

foff = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 10, 'format': '10d'})

grn = Column(Integer,
             info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

grname = Column(String(40),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a40', 'width': 40, 'format': '40.40s'})

hang = Column(Float(24),
              info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.1f'})

inarrival = Column(String(1),
                   info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

inid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

insname = Column(String(50),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})

instant = Column(String(1),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

instype = Column(String(6),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

iphase = Column(String(8),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

jdate = Column(Integer,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

keyname = Column(String(15),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

keyvalue = Column(Integer,
                  info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

lat = Column(Float(53),
             info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 11, 'format': '11.6f'})

lddate = Column(DateTime, nullable=False, onupdate=datetime.now,
                info={'default': datetime.now, 'parse': dtfn, 'dtype': 'O', 'width': 19, 'format': DATEFMT})

lineno = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

logat = Column(Float(24),
               info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

lon = Column(Float(53),
             info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 11, 'format': '11.6f'})

magid = Column(Integer, nullable=False,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

magnitude = Column(Float(24),
                   info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

magres = Column(Float(24),
                info={'default': -999, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

magdef = Column(String(1),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

magtype = Column(String(6),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

mb = Column(Float(24),
            info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

mbid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

ml = Column(Float(24),
            info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

mlid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

ms = Column(Float(24),
            info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

msid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

mmodel = Column(String(15),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

nass = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})

ncalib = Column(Float(24),
                info={'default': 1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

ncalper = Column(Float(24),
                 info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})

ndef = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})

ndp = Column(Integer,
             info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})

net = Column(String(8),
             info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

netname = Column(String(80),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a80', 'width': 80, 'format': '80.80s'})

nettype = Column(String(4),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})

nsamp = Column(Integer,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

nsta = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

offdate = Column(Integer,
                 info={'default': 2286324, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

ondate = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

orid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

parid = Column(Integer,
               info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

per = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

phase = Column(String(8),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

prefor = Column(Integer,
                info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '9d'})

qual = Column(String(1),
              info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

rect = Column(Float(24),
              info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.3f'})

refsta = Column(String(6),
                info={'default': '-', 'parse': parse_str, 'dtype': 'str', 'width': 6, 'format': '6.6s'})

remark = Column(String(80),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a80', 'width': 80, 'format': '80.80s'})

rsptype = Column(String(6),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

samprate = Column(Float(24),
                  info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 11, 'format': '11.7f'})

sdepth = Column(Float(24),
                info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

sdobs = Column(Float(24),
               info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

seaz = Column(Float(24),
              info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

segtype = Column(String(1),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

slodef = Column(String(1),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

slores = Column(Float(24),
                info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

slow = Column(Float(24),
              info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

smajax = Column(Float(24),
                info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

sminax = Column(Float(24),
                info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})

snr = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 10, 'format': '10.2f'})

srn = Column(Integer,
             info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})

srname = Column(String(80),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a80', 'width': 80, 'format': '80.80s'})

sta = Column(String(6),
             info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})

staname = Column(String(50),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})

stassid = Column(Integer,
                 info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

statype = Column(String(4),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})

stime = Column(Float(24),
               info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.2f'})

strike = Column(Float(24),
                info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.2f'})

stt = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

stx = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

sty = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

stype = Column(String(1),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

stz = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

sxx = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

sxy = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

sxz = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

syy = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

syz = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

szz = Column(Float(24),
             info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})

tagid = Column(Integer,
               info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9d'})

tagname = Column(String(8),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})

time = Column(Float(53),
              info={'default': -9999999999.999, 'parse': parse_float, 'dtype': 'float', 'width': 17, 'format': '17.5f'})

timedef = Column(String(1),
                 info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})

timeres = Column(Float(24),
                 info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8.3f'})

tshift = Column(Float(24),
                info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.2f'})

uncertainty = Column(Float(24),
                     info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})

units = Column(String(15),
               info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

vang = Column(Float(24),
              info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.1f'})

vmodel = Column(String(15),
                info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})

wfid = Column(Integer,
              info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 9, 'format': '9d'})

wgt = Column(Float(24),
             info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.3f'})


class Affiliation(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('net', 'sta', 'time'),)

    net = net.copy()
    sta = sta.copy()
    time = time.copy()
    endtime = endtime.copy()
    lddate = lddate.copy()


class Amplitude(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('ampid'),)

    ampid = ampid.copy()
    arid = arid.copy()
    parid = parid.copy()
    chan = chan.copy()
    amp = amp.copy()
    per = per.copy()
    snr = snr.copy()
    amptime = amptime.copy()
    time = time.copy()
    duration = duration.copy()
    deltaf = deltaf.copy()
    amptype = amptype.copy()
    units = units.copy()
    clip = clip.copy()
    inarrival = inarrival.copy()
    auth = auth.copy()
    lddate = lddate.copy()


class Arrival(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('arid'),
                UniqueConstraint('sta', 'time', 'chan', 'iphase', 'auth'),)

    sta = sta.copy()
    time = time.copy()
    arid = arid.copy()
    jdate = jdate.copy()
    stassid = stassid.copy()
    chanid = chanid.copy()
    chan = chan.copy()
    iphase = iphase.copy()
    stype = stype.copy()
    deltim = deltim.copy()
    azimuth = azimuth.copy()
    delaz = delaz.copy()
    slow = slow.copy()
    delslo = delslo.copy()
    ema = ema.copy()
    rect = rect.copy()
    amp = amp.copy()
    per = per.copy()
    logat = logat.copy()
    clip = clip.copy()
    fm = fm.copy()
    snr = snr.copy()
    qual = qual.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Assoc(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('arid', 'orid'), UniqueConstraint('arid'),)

    arid = arid.copy()
    orid = orid.copy()
    sta = sta.copy()
    phase = phase.copy()
    belief = belief.copy()
    delta = delta.copy()
    seaz = seaz.copy()
    esaz = esaz.copy()
    timeres = timeres.copy()
    timedef = timedef.copy()
    azres = azres.copy()
    azdef = azdef.copy()
    slores = slores.copy()
    slodef = slodef.copy()
    emares = emares.copy()
    wgt = wgt.copy()
    vmodel = vmodel.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Event(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('evid'), UniqueConstraint('prefor'),)
    evid = evid.copy()
    evname = evname.copy()
    prefor = prefor.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Gregion(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('grn'),)

    grn = grn.copy()
    grname = grname.copy()
    lddate = lddate.copy()


class Instrument(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('inid'),)

    inid = inid.copy()
    insname = insname.copy()
    instype = instype.copy()
    band = band.copy()
    digital = digital.copy()
    samprate = samprate.copy()
    ncalib = ncalib.copy()
    ncalper = ncalper.copy()
    dir = dir.copy()
    dfile = dfile.copy()
    rsptype = rsptype.copy()
    lddate = lddate.copy()


# TODO: implement class-wide generator expressions on keyvalues.
#   https://groups.google.com/forum/#!topic/sqlalchemy/XksPIVYOdSU
# or something else
# http://stackoverflow.com/questions/10494033/setting-sqlalchemy-autoincrement-start-value
# @attr_generate
class Lastid(Base):
    """
    Store last id values.

    """
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('keyname'), UniqueConstraint('keyname', 'keyvalue'),)

    def __next__(self):
        self.keyvalue += 1
        return self.keyvalue

    keyname = keyname.copy()
    keyvalue = keyvalue.copy()
    lddate = lddate.copy()


class Netmag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('magid'),
                UniqueConstraint('magid', 'orid'),)

    magid = magid.copy()
    net = net.copy()
    orid = orid.copy()
    evid = evid.copy()
    magtype = magtype.copy()
    nsta = nsta.copy()
    magnitude = magnitude.copy()
    uncertainty = uncertainty.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Network(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('net'),)

    net = net.copy()
    netname = netname.copy()
    nettype = nettype.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Origerr(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('orid'),)

    orid = orid.copy()
    sxx = sxx.copy()
    syy = syy.copy()
    szz = szz.copy()
    stt = stt.copy()
    sxy = sxy.copy()
    sxz = sxz.copy()
    syz = syz.copy()
    stx = stx.copy()
    sty = sty.copy()
    stz = stz.copy()
    sdobs = sdobs.copy()
    smajax = smajax.copy()
    sminax = sminax.copy()
    strike = strike.copy()
    sdepth = sdepth.copy()
    stime = stime.copy()
    conf = conf.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Origin(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('lat', 'lon', 'depth', 'time', 'auth'), PrimaryKeyConstraint('orid'),)

    lat = lat.copy()
    lon = lon.copy()
    depth = depth.copy()
    time = time.copy()
    orid = orid.copy()
    evid = evid.copy()
    jdate = jdate.copy()
    nass = nass.copy()
    ndef = ndef.copy()
    ndp = ndp.copy()
    grn = grn.copy()
    srn = srn.copy()
    etype = etype.copy()
    depdp = depdp.copy()
    dtype = dtype.copy()
    mb = mb.copy()
    mbid = mbid.copy()
    ms = ms.copy()
    msid = msid.copy()
    ml = ml.copy()
    mlid = mlid.copy()
    algorithm = algorithm.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Remark(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('commid', 'lineno'),)

    commid = commid.copy()
    lineno = lineno.copy()
    remark = remark.copy()
    lddate = lddate.copy()


class Sensor(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('sta', 'chan', 'time', 'endtime'),)

    sta = sta.copy()
    chan = chan.copy()
    time = time.copy()
    endtime = endtime.copy()
    inid = inid.copy()
    chanid = chanid.copy()
    jdate = jdate.copy()
    calratio = calratio.copy()
    calper = calper.copy()
    tshift = tshift.copy()
    instant = instant.copy()
    lddate = lddate.copy()


class Site(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('sta', 'ondate'),)

    sta = sta.copy()
    ondate = ondate.copy()
    offdate = offdate.copy()
    lat = lat.copy()
    lon = lon.copy()
    elev = elev.copy()
    staname = staname.copy()
    statype = statype.copy()
    refsta = refsta.copy()
    dnorth = dnorth.copy()
    deast = deast.copy()
    lddate = lddate.copy()


class Sitechan(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('sta', 'chan', 'ondate'),
                PrimaryKeyConstraint('chanid'))

    sta = sta.copy()
    chan = chan.copy()
    ondate = ondate.copy()
    chanid = chanid.copy()
    offdate = offdate.copy()
    ctype = ctype.copy()
    edepth = edepth.copy()
    hang = hang.copy()
    vang = vang.copy()
    descrip = descrip.copy()
    lddate = lddate.copy()


class Sregion(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('srn'),)

    srn = srn.copy()
    srname = srname.copy()
    lddate = lddate.copy()


class Stamag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('magid', 'sta', 'arid'),)

    magid = magid.copy()
    ampid = ampid.copy()
    sta = sta.copy()
    arid = arid.copy()
    orid = orid.copy()
    evid = evid.copy()
    phase = phase.copy()
    delta = delta.copy()
    magtype = magtype.copy()
    magnitude = magnitude.copy()
    uncertainty = uncertainty.copy()
    magres = magres.copy()
    magdef = magdef.copy()
    mmodel = mmodel.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Wfdisc(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('wfid', 'dir', 'dfile'), PrimaryKeyConstraint('wfid'),)

    def to_trace(self):
        """
        Read the wfdisc line into a Trace instance.  Minimal header.

        Returns
        -------
        obspy.Trace

        """
        return wfdisc2trace(self)

    sta = sta.copy()
    chan = chan.copy()
    time = time.copy()
    wfid = wfid.copy()
    chanid = chanid.copy()
    jdate = jdate.copy()
    endtime = endtime.copy()
    nsamp = nsamp.copy()
    samprate = samprate.copy()
    calib = calib.copy()
    calper = calper.copy()
    instype = instype.copy()
    segtype = segtype.copy()
    datatype = datatype.copy()
    clip = clip.copy()
    dir = dir.copy()
    dfile = dfile.copy()
    foff = foff.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Wftag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('tagid'), UniqueConstraint('tagname', 'tagid', 'wfid'),)

    tagname = tagname.copy()
    tagid = tagid.copy()
    wfid = wfid.copy()
    lddate = lddate.copy()


# possibly useful indexes
# To add, use Index.create(engine)
# ix_affiliation_stanet = Index(u'ix_affiliation_stanet',
#                              Affiliation.__table__.c.sta,
#                              Affiliation.__table__.c.net)
#
# ix_amplitude_uk = Index(u'ix_amplitude_uk',
#                        Amplitude.__table__.c.arid,
#                        Amplitude.__table__.c.amptime,
#                        Amplitude.__table__.c.amptype,
#                        Amplitude.__table__.c.auth,
#                        Amplitude.__table__.c.chan,
#                        Amplitude.__table__.c.deltaf,
#                        Amplitude.__table__.c.duration,
#                        Amplitude.__table__.c.parid,
#                        Amplitude.__table__.c.per,
#                        Amplitude.__table__.c.time)
#
# ix_arrival_uk = Index(u'ix_arrival_uk',
#                      Arrival.__table__.c.time,
#                      Arrival.__table__.c.sta,
#                      Arrival.__table__.c.chan,
#                      Arrival.__table__.c.iphase,
#                      Arrival.__table__.c.auth)
#
# ix_event_evid_prefor = Index(u'ix_event_evid_prefor',
#                             Event.__table__.c.evid,
#                             Event.__table__.c.prefor)
#
# ix_instrument_uk = Index(u'ix_instrument_uk',
#                         Instrument.__table__.c.dfile,
#                         Instrument.__table__.c.dir,
#                         Instrument.__table__.c.instype,
#                         Instrument.__table__.c.ncalib,
#                         Instrument.__table__.c.samprate)
#
# ix_netmaguk = Index(u'ix_netmaguk',
#                    Netmag.__table__.c.orid,
#                    Netmag.__table__.c.magtype,
#                    Netmag.__table__.c.auth)
#
# ix_network_uk = Index(u'ix_network_uk',
#                      Network.__table__.c.auth,
#                      Network.__table__.c.netname)
#
# ix_originautheviduk = Index(u'ix_originautheviduk',
#                            Origin.__table__.c.evid,
#                            Origin.__table__.c.auth)
#
# ix_origin_uk = Index(u'origin_uk',
#                     Origin.__table__.c.lat,
#                     Origin.__table__.c.lon,
#                     Origin.__table__.c.depth,
#                     Origin.__table__.c.time,
#                     Origin.__table__.c.auth)
#
# ix_b_newsite_onoff_ix = Index(u'ix_b_newsite_onoff',
#                              Site.__table__.c.ondate,
#                              Site.__table__.c.offdate)
#
# ix_site = Index(u'ix_site',
#                Site.__table__.c.sta,
#                Site.__table__.c.ondate,
#                Site.__table__.c.offdate)
#
# ix_sitechan_uk = Index(u'ix_sitechan_uk',
#                       Sitechan.__table__.c.sta,
#                       Sitechan.__table__.c.chan,
#                       Sitechan.__table__.c.ondate)
#
# ix_wfdisc_dirdfile = Index(u'ix_wfdisc_dirdfile',
#                           Wfdisc.__table__.c.dfile,
#                           Wfdisc.__table__.c.dir,
#                           Wfdisc.__table__.c.foff)
#
# ix_wfdisc = Index(u'ix_wfdisc',
#                  Wfdisc.__table__.c.sta,
#                  Wfdisc.__table__.c.jdate,
#                  Wfdisc.__table__.c.time,
#                  Wfdisc.__table__.c.endtime,
#                  Wfdisc.__table__.c.wfid)
#
# ix_wfdisc_uk = Index(u'ix_wfdisc_uk',
#                     Wfdisc.__table__.c.sta,
#                     Wfdisc.__table__.c.chan,
#                     Wfdisc.__table__.c.time)
#
# ix_wfdisc_chanid_instype = Index(u'ix_wfdisc_chanid_instype',
#                                 Wfdisc.__table__.c.chanid,
#                                 Wfdisc.__table__.c.instype)



from datetime import datetime
import numpy as np

from obspy.core import UTCDateTime
from pisces.schema.util import get_infovals


# TODO: replace numpy.genfromtxt with pandas.read_fwf ?

def format_records(recs, meta, structure):
    """
    Get a well-formatted rec.array of records with requested structure.

    Returned array has the desired field order with any missing values.

    recs: numpy structured array
    meta: sqlalchemy.MetaData 
        Columns in tables must have info['dtype'] and info['width']
    structure: iterable of str
        The structure of the text file.  Strings are known schema objects 
        (tables, columns).  Duplicate column names will be overwritten!


    """
    # The Plan:
    # 1) Make 2 new intermediate record arrays.  1 contains missing values for
    #    all expected columns not in the incoming array, the other is empty and
    #    has the requested output structure.
    # 2) Fill in the empty recarray with either values from the incoming array,
    #    or, if not present, pull from the missing value array. Return it.

    # get flat expected field names and defaults from structure
    fields, dflts = get_infovals(meta, structure, 'default')
    mdict = dict(zip(fields, dflts))

    # initialize output array
    fields, dtypes = get_infovals(meta, structure, 'dtype')
    dtp = np.dtype(zip(fields, dtypes))
    newrecs = np.recarray(recs.shape, dtype=dtp)

    # fill in any missing values
    for field in fields:
        try:
            newrecs[field] = recs[field]
        except ValueError:
            #field name not found in recs.  must be a missing value
            newrecs[field] = mdict[field]

    return newrecs



def read_flatfile(file_ish, meta, structure):
    """
    Read formatted flat file with known schema objects.

    Parameters
    ----------
    file_ish: str of file-like object
        Fixed-width text file.
    meta: sqlalchemy.MetaData 
        Columns in tables must have info['dtype'] and info['width']
    structure: iterable of str
        The structure of the text file.  Strings are known schema objects 
        (tables, columns).  Duplicate column names will be overwritten!
    
    Returns
    -------
    numpy.rec.array of np.records

    Notes
    -----


    """
    isFileString = isinstance(file_ish, str)

    #prepare for numpy.genfromtxt
    fields, dtypes = get_infovals(meta, structure, 'dtype')
    dtp = np.dtype(zip(fields, dtypes))

    fields, widths = get_infovals(meta, structure, 'width')
    widths = np.array(widths) + 1
    widths[-1] -= 1

    # read the file into a record array
    # try a series of lddate converters
    #1) Anything UTCDateTime can chomp on (which is a lot)
    #2) e.g. '07-JUL-10'
    # TODO: this is kinda hackish
    for conv in [lambda x: UTCDateTime(x).datetime, 
                 lambda x: UTCDateTime().strptime(x, '%d-%b-%y').datetime]:
        try:
            rec = np.genfromtxt(file_ish, usemask=False, autostrip=True, 
                    names=fields, dtype=dtp, delimiter=widths, 
                    converters={'lddate': conv})
            break
        except (TypeError, ValueError) as e:
            #wrong lddate converter
            pass

    #make sure you return a numpy.recarray of records (i.e. type(rec[0]) is numpy.rec.record)
    #rec = np.rec.array(rec.copy()) #works with genfromtxt, unless fields are objects (e.g. datetime)
    rec = format_records(rec, meta, structure) 
    
    return rec





# coding: utf-8
"""
Antelope v4.11 Datascope database schema


"""
from datetime import datetime
from sqlalchemy import Date, DateTime, Float, Numeric, String, Integer
from sqlalchemy import Column, Table, func
from sqlalchemy import PrimaryKeyConstraint, UniqueConstraint, Index
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import declared_attr

from obspy.core import UTCDateTime
from pisces.schema.util import PiscesMeta
from pisces.schema.util import parse_int, parse_float, parse_str

from pisces.io.trace import wfdisc2trace

Base = declarative_base(metaclass=PiscesMeta, constructor=None)

# COLUMN DEFINITIONS
# Generic SQLA types, compatible with different backends
# !! info dictionary defines the external representations (NumPy, text files)
#   and default values for the mapped class representation.
# XXX: for numeric types, maximum width is not enforced!
# NOTE: info['dtype'] for floats/ints should match the system default for Python
#       "{:f}".format(numpyfloat/int) to work, b/c it can use the builtin float
#       __format__().  use 'float' or 'int'
#       See:
#       http://stackoverflow.com/questions/16928644/floats-in-numpy-structured-array-and-native-string-formatting-with-format

def strip(s):
    return str(s).strip()

DATEFMT = '%y-%m-%d %H:%M:%S'
def dtfn(s):
    try:
        dt = datetime.strptime(s.strip(), DATEFMT)
    except ValueError:
        # Antelope convention
        dt = UTCDateTime(float(s)).datetime
    return dt


algorithm = Column(String(15),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})
amp = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 10, 'format': '10.1f'})
ampid = Column(Integer, nullable=False,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
amptime = Column(Float(53), info={'default': -9999999999.999,
        'parse': parse_float, 'dtype': 'float', 'width': 17, 'format': '17.5f'})
amptype = Column(String(8),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})
arid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
auth = Column(String(15),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15s'})
azdef = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
azimuth = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
azres = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.1f'})
band = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
belief = Column(Float(24),
        info={'default': -9.99, 'parse': parse_float, 'dtype': 'float', 'width': 4, 'format': '4.2f'})
calib = Column(Float(24),
        info={'default': 0.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.9f'})
calper = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})
calratio = Column(Float(24),
        info={'default': 1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})
chan = Column(String(8),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})
chanid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
clip = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
commid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
conf = Column(Float(24),
        info={'default': 0, 'parse': parse_float, 'dtype': 'float', 'width': 5, 'format': '5.3f'})
ctype = Column(String(4),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})
datatype = Column(String(2),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a2', 'width': 2, 'format': '2.2s'})
deast = Column(Float(24),
        info={'default': 0.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
delaz = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
delslo = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
delta = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8.3f'})
deltaf = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.3f'})
deltim = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.3f'})
depdp = Column(Float(24),
        info={'default': -999, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
depth = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
descrip = Column(String(50),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})
dfile = Column(String(32),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a32', 'width': 32, 'format': '32.32s'})
digital = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
dir = Column(String(64),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a64', 'width': 64, 'format': '64.64s'})
dnorth = Column(Float(24),
        info={'default': 0.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
dtype = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
duration = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
edepth = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
elev = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
ema = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
emares = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.1f'})
endtime = Column(Float(53), info={'default':  9999999999.999,
        'parse': parse_float, 'dtype': 'float', 'width': 17, 'format': '17.5f'})
esaz = Column(Float(24),
        info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
etype = Column(String(2),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a2', 'width': 2, 'format': '2.2s'})
evid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
evname = Column(String(15),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})
fm = Column(String(2),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a2', 'width': 2, 'format': '2.2s'})
foff = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 10, 'format': '10d'})
grn = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
grname = Column(String(40),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a40', 'width': 40, 'format': '40.40s'})
hang = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.1f'})
inarrival = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
inid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
insname = Column(String(50),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})
instant = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
instype = Column(String(6),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})
iphase = Column(String(8),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})
jdate = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
keyname = Column(String(15),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})
keyvalue = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
lat = Column(Float(53),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
lddate = Column(DateTime, nullable=False, onupdate=datetime.now,
        info={'default': datetime.now, 'parse': dtfn, 'dtype': 'O', 'width': 17, 'format': DATEFMT})
lineno = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
logat = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
lon = Column(Float(53),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
magid = Column(Integer, nullable=False,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
magnitude = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
magres = Column(Float(24),
        info={'default': -999, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
magdef = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
magtype = Column(String(6),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})
mb = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
mbid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
ml = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
mlid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
ms = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
msid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
mmodel = Column(String(15),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})
nass = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})
ncalib = Column(Float(24),
        info={'default': 1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})
ncalper = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 16, 'format': '16.6f'})
ndef = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})
ndp = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 4, 'format': '4d'})
net = Column(String(8),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})
netname = Column(String(80),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a80', 'width': 80, 'format': '80.80s'})
nettype = Column(String(4),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})
nsamp = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
nsta = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
offdate = Column(Integer,
        info={'default': 2286324, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
ondate = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
orid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
parid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
per = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
phase = Column(String(8),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})
prefor = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
qual = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
rect = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.3f'})
refsta = Column(String(6),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})
remark = Column(String(80),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a80', 'width': 80, 'format': '80.80s'})
review =  Column(String(4),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})
rsptype = Column(String(6),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})
samprate = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 11, 'format': '11.7f'})
sdepth = Column(Float(24),
        info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
sdobs = Column(Float(24),
        info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
seaz = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
segtype = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
slodef = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
slores = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
slow = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float','width': 7, 'format': '7.2f'})
smajax = Column(Float(24),
        info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
sminax = Column(Float(24),
        info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 9, 'format': '9.4f'})
snr = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 10, 'format': '10.5f'})
srn = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
srname = Column(String(40),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a40', 'width': 40, 'format': '40.40s'})
sta = Column(String(6),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a6', 'width': 6, 'format': '6.6s'})
staname = Column(String(50),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a50', 'width': 50, 'format': '50.50s'})
stassid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
statype = Column(String(4),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a4', 'width': 4, 'format': '4.4s'})
stime = Column(Float(24),
        info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8.2f'})
strike = Column(Float(24),
        info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.2f'})
stt = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
stx = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
sty = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
stype = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
stz = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
sxx = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
sxy = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
sxz = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
syy = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
syz = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
szz = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 15, 'format': '15.4f'})
tagid = Column(Integer,
        info={'default': -1, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8d'})
tagname = Column(String(8),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a8', 'width': 8, 'format': '8.8s'})
time = Column(Float(53),
        info={'default': -9999999999.999, 'parse': parse_float, 'dtype': 'float', 'width': 17, 'format': '17.5f'})
timedef = Column(String(1),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a1', 'width': 1, 'format': '1.1s'})
timeres = Column(Float(24),
        info={'default': -999.0, 'parse': parse_float, 'dtype': 'float', 'width': 8, 'format': '8.3f'})
tshift = Column(Float(24),
        info={'default': -100000000, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.2f'})
uncertainty = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 7, 'format': '7.2f'})
units = Column(String(12),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a12', 'width': 12, 'format': '12.12s'})
vang = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.1f'})
vmodel = Column(String(15),
        info={'default': '-', 'parse': parse_str, 'dtype': 'a15', 'width': 15, 'format': '15.15s'})
wfid = Column(Integer,
        info={'default': -1, 'parse': parse_int, 'dtype': 'int', 'width': 8, 'format': '8d'})
wgt = Column(Float(24),
        info={'default': -1.0, 'parse': parse_float, 'dtype': 'float', 'width': 6, 'format': '6.3f'})

class Affiliation(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('net', 'sta', 'time'),)

    net = net.copy()
    sta = sta.copy()
    lddate = lddate.copy()


class Amplitude(Base):
    __abstract__ =True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('ampid'),)

    ampid = ampid.copy()
    arid = arid.copy()
    parid = parid.copy()
    chan = chan.copy()
    amp = amp.copy()
    per = per.copy()
    snr = snr.copy()
    amptime = amptime.copy()
    time = time.copy()
    duration = duration.copy()
    deltaf = deltaf.copy()
    amptype = amptype.copy()
    units = units.copy()
    clip = clip.copy()
    inarrival = inarrival.copy()
    auth = auth.copy()
    lddate = lddate.copy()


class Arrival(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('arid'),
                UniqueConstraint(u'sta', u'time', u'chan', u'iphase', u'auth'),)

    sta = sta.copy()
    time = time.copy()
    arid = arid.copy()
    jdate = jdate.copy()
    stassid = stassid.copy()
    chanid = chanid.copy()
    chan = chan.copy()
    iphase = iphase.copy()
    stype = stype.copy()
    deltim = deltim.copy()
    azimuth = azimuth.copy()
    delaz = delaz.copy()
    slow = slow.copy()
    delslo = delslo.copy()
    ema = ema.copy()
    rect = rect.copy()
    amp = amp.copy()
    per = per.copy()
    logat = logat.copy()
    clip = clip.copy()
    fm = fm.copy()
    snr = snr.copy()
    qual = qual.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Assoc(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('arid','orid'),
                UniqueConstraint('arid'),)

    arid = arid.copy()
    orid = orid.copy()
    sta = sta.copy()
    phase = phase.copy()
    belief = belief.copy()
    delta = delta.copy()
    seaz = seaz.copy()
    esaz = esaz.copy()
    timeres = timeres.copy()
    timedef = timedef.copy()
    azres = azres.copy()
    azdef = azdef.copy()
    slores = slores.copy()
    slodef = slodef.copy()
    emares = emares.copy()
    wgt = wgt.copy()
    vmodel = vmodel.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Event(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('evid'),
                UniqueConstraint('prefor'),)

    evid = evid.copy()
    evname = evname.copy()
    prefor = prefor.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Gregion(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('grn'),)

    grn = grn.copy()
    grname = grname.copy()
    lddate = lddate.copy()


class Instrument(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('inid'),)

    inid = inid.copy()
    insname = insname.copy()
    instype = instype.copy()
    band = band.copy()
    digital = digital.copy()
    samprate = samprate.copy()
    ncalib = ncalib.copy()
    ncalper = ncalper.copy()
    dir = dir.copy()
    dfile = dfile.copy()
    rsptype = rsptype.copy()
    lddate = lddate.copy()


class Lastid(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('keyname'),
                UniqueConstraint(u'keyname', u'keyvalue'),)

    keyname = keyname.copy()
    keyvalue = keyvalue.copy()
    lddate = lddate.copy()


class Netmag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('magid'),
                UniqueConstraint('magid', 'orid'),)

    magid = magid.copy()
    net = net.copy()
    orid = orid.copy()
    evid = evid.copy()
    magtype = magtype.copy()
    nsta = nsta.copy()
    magnitude = magnitude.copy()
    uncertainty = uncertainty.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Network(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('net'),)

    net = net.copy()
    netname = netname.copy()
    nettype = nettype.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Origerr(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('orid'),)

    orid = orid.copy()
    sxx = sxx.copy()
    syy = syy.copy()
    szz = szz.copy()
    stt = stt.copy()
    sxy = sxy.copy()
    sxz = sxz.copy()
    syz = syz.copy()
    stx = stx.copy()
    sty = sty.copy()
    stz = stz.copy()
    sdobs = sdobs.copy()
    smajax = smajax.copy()
    sminax = sminax.copy()
    strike = strike.copy()
    sdepth = sdepth.copy()
    stime = stime.copy()
    conf = conf.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Origin(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('lat','lon','depth','time','auth'),
                PrimaryKeyConstraint('orid'),)

    lat = lat.copy()
    lon = lon.copy()
    depth = depth.copy()
    time = time.copy()
    orid = orid.copy()
    evid = evid.copy()
    jdate = jdate.copy()
    nass = nass.copy()
    ndef = ndef.copy()
    ndp = ndp.copy()
    grn = grn.copy()
    srn = srn.copy()
    etype = etype.copy()
    review = review.copy()
    depdp = depdp.copy()
    dtype = dtype.copy()
    mb = mb.copy()
    mbid = mbid.copy()
    ms = ms.copy()
    msid = msid.copy()
    ml = ml.copy()
    mlid = mlid.copy()
    algorithm = algorithm.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Remark(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('commid','lineno'),)

    commid = commid.copy()
    lineno = lineno.copy()
    remark = remark.copy()
    lddate = lddate.copy()


class Sensor(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('sta','chan','time','endtime'),)

    sta = sta.copy()
    chan = chan.copy()
    time = time.copy()
    endtime = endtime.copy()
    inid = inid.copy()
    chanid = chanid.copy()
    jdate = jdate.copy()
    calratio = calratio.copy()
    calper = calper.copy()
    tshift = tshift.copy()
    instant = instant.copy()
    lddate = lddate.copy()


class Site(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('sta','ondate'),)

    sta = sta.copy()
    ondate = ondate.copy()
    offdate = offdate.copy()
    lat = lat.copy()
    lon = lon.copy()
    elev = elev.copy()
    staname = staname.copy()
    statype = statype.copy()
    refsta = refsta.copy()
    dnorth = dnorth.copy()
    deast = deast.copy()
    lddate = lddate.copy()


class Sitechan(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('sta','chan','ondate'),
                PrimaryKeyConstraint('sta','chan','ondate','offtime'),)

    sta = sta.copy()
    chan = chan.copy()
    ondate = ondate.copy()
    chanid = chanid.copy()
    offdate = offdate.copy()
    ctype = ctype.copy()
    edepth = edepth.copy()
    hang = hang.copy()
    vang = vang.copy()
    descrip = descrip.copy()
    lddate = lddate.copy()


class Sregion(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('srn'),)

    srn = srn.copy()
    srname = srname.copy()
    lddate = lddate.copy()


class Stamag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('magid', 'sta', 'arid'),)

    magid = magid.copy()
    sta = sta.copy()
    arid = arid.copy()
    orid = orid.copy()
    evid = evid.copy()
    phase = phase.copy()
    magtype = magtype.copy()
    magnitude = magnitude.copy()
    uncertainty = uncertainty.copy()
    auth = auth.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Wfdisc(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (UniqueConstraint('wfid','dir','dfile'),
                PrimaryKeyConstraint('wfid'),)

    def to_trace(self):
         """
         Read the wfdisc line into a Trace instance.  Minimal header.

         Returns
         -------
         obspy.Trace

         """
         return wfdisc2trace(self)

    sta = sta.copy()
    chan = chan.copy()
    time = time.copy()
    wfid = wfid.copy()
    chanid = chanid.copy()
    jdate = jdate.copy()
    endtime = endtime.copy()
    nsamp = nsamp.copy()
    samprate = samprate.copy()
    calib = calib.copy()
    calper = calper.copy()
    instype = instype.copy()
    segtype = segtype.copy()
    datatype = datatype.copy()
    clip = clip.copy()
    dir = dir.copy()
    dfile = dfile.copy()
    foff = foff.copy()
    commid = commid.copy()
    lddate = lddate.copy()


class Wftag(Base):
    __abstract__ = True

    @declared_attr
    def __table_args__(cls):
        return (PrimaryKeyConstraint('tagid'),
                UniqueConstraint('tagname','tagid','wfid'),)

    tagname = tagname.copy()
    tagid = tagid.copy()
    wfid = wfid.copy()
    lddate = lddate.copy()





# Converted to python 3.5.2 on 02/07/17
# By: Jeremy Webster

"""
Home to read_waveform and all format-specific reading functions.

"""
from distutils import sysconfig
import ctypes as C
import os
from io import StringIO

import numpy as np
from obspy.core import read

# C libraries
ext, = sysconfig.get_config_vars('SO')
libecomp = C.CDLL(os.path.dirname(__file__) + '/lib/libecompression' + ext)
libconvert = C.CDLL(os.path.dirname(__file__) + '/lib/libconvert' + ext)

# TODO: implement other datatypes

# NumPy-supported datatypes {datatype: numpy.dtype}
NPTYPE = {'t4': '>f4', 't8': '>f8', 's4': '>i4', 's2': '>i2', 's1': '>i1',
          'f8': '<f8', 'f4': '<f4', 'i4': '<i4', 'i2': '<i2', 'i1': '<i1',
          'p8': '<f8', 'p4': '<f4', 't4': '>f4', 'u4': '<f4'}

# ObsPy-supported datatypes
OBSPYTYPE = {'sd': 'MSEED', 'sc': 'SAC', 'sy': 'SEGY'}
OBSPYTYPE = {'sc': 'SAC', 'sy': 'SEGY'}
# NOTE: 'sd' can also mean full SEED.


def read_waveform(DATAFILE, DATATYPE, BYTEOFFSET, NUM):
    """
    Read a data vector of a certain data type.

    Parameters
    ----------
    DATAFILE : string
        Path and name of data file.
    DATATYPE : string
        {'f4', 'i1', 'i2', 'i4', 'p4', 't8', 's2', 'p8', 's4', 'f8', 'u4', 't4',
         's1', 'sd', 'e1', 'sc', 'sy'}
        Data type string (from wfdisc.datatype).
        'sd', 'sc', and 'sy' are SEED/miniSEED, SAC, and SEGY
    BYTEOFFSET : int
        Number of bytes from beginning of file.
    NUM : int
        Number of samples to read.

    Returns
    -------
    data : numpy.array
        Vector of NUM samples of type DATATYPE.

    Raises
    ------
    ValueError
        Unrecognized DATATYPE.
    IOError
        DATAFILE does not exist.

    Notes
    -----
    If the file has a header and you want it, you'll have to read it manually
    with obspy.core.read or something else.

    """
    # TODO: do something with return flags
    if DATATYPE in NPTYPE:
        data = numpy_read(DATAFILE, BYTEOFFSET, NUM, 'rb', NPTYPE[DATATYPE])
    elif DATATYPE in OBSPYTYPE:
        data = read(DATAFILE, format=OBSPYTYPE[DATATYPE])[0].data
    elif DATATYPE == 'sd':
        # this datatype is not well-suited for a wfdisc, as it supports
        # multiplexing and wfdisc does not
        data = read_seed(DATAFILE, BYTEOFFSET)
    elif DATATYPE == 'e1':
        # data = read_e1(DATAFILE, BYTEOFFSET, NUM)
        data = e_compression(DATAFILE, BYTEOFFSET, NUM)
    elif DATATYPE == 's3':
        data, flag = read_s3(DATAFILE, BYTEOFFSET, NUM)
    else:
        raise ValueError("Unrecognized data format: {0}".format(DATATYPE))

    # if len(data) not NUM:
    #    # TODO: raise something here
    #    pass

    return data


def e_compression(DATAFILE, BYTEOFFSET, NUM):
    """Wrapper to e1 decompression routine.

    Parameters
    ----------
    DATAFILE: string
        Full path to e1 file.
    BYTEOFFSET: int
        Number of bytes to start of the data.
    NUM: int
        Number of expected samples.

    Returns
    -------
    data: numpy.array (rank 1) of int32
        Uncompressed data vector.
    retval: int
        Return value/code. One of the following:

        * 0: 'EC_SUCCESS',
        * 1: 'EC_FAILED',
        * 2: 'EC_LENGTH_ERROR',
        * 3: 'EC_SAMP_ERROR',
        * 4: 'EC_DIFF_ERROR',
        * 5: 'EC_CHECK_ERROR',
        * 6: 'EC_ARG_ERROR',
        * 7: 'EC_TYPE_ERROR',
        * 8: 'EC_MEMORY_ERROR'

    Raises
    ------
    ValueError : BYTEOFFSET exceeds file size.

    Notes
    -----
    e1 decompression C library is written by Richard Stead, LANL.

    """
    # int32_t
    # e_decomp_inplace(int32_t *in, int32_t insamp, int32_t inbyte, int32_t out0,
    #   int32_t outsamp)
    # libecomp.e_decomp_inplace.restype = C.c_int

    # int32_t
    # e_decomp(uint32_t *in, int32_t *out, int32_t insamp, int32_t inbyte,
    #  int32_t out0, int32_t outsamp)
    libecomp.e_decomp.restype = C.c_int

    # open file, query size, jump to offset
    f = open(DATAFILE, 'rb')
    flen = os.stat(DATAFILE).st_size
    if flen < BYTEOFFSET:
        raise ValueError("BYTEOFFSET exceeds file size.")
    f.seek(BYTEOFFSET)
    flen -= BYTEOFFSET
    # flen is capped at 5*NUM
    flen = 5 * NUM if flen > 5 * NUM else flen

    # Read the entire compressed chunk
    # w = (int32_t *)malloc((int)((flen + 3) / 4) * 4)
    # read(fd, w, flen)
    w = np.fromfile(f, count=flen, dtype=np.int32)  # XXX: check this
    f.close()

    Y = np.zeros(NUM, dtype=np.int32, order='C')

    # library call
    # replaces the first NUM values in w with decompressed values
    # retval = libecomp.e_decomp_inplace(w.ctypes.data_as(C.POINTER(C.c_int32)),
    #                                   NUM, flen, 0, NUM)
    retval = libecomp.e_decomp(w.ctypes.data_as(C.POINTER(C.c_uint32)),
                               Y.ctypes.data_as(C.POINTER(C.c_int32)), NUM,
                               flen, 0, NUM)

    # return w[0:NUM].copy(), retval
    # return w.copy(), retval
    return Y


def read_seed(DATAFILE, BYTEOFFSET):
    """Read a SEED or miniSEED 'sd' datatype.
    """
    if BYTEOFFSET:
        with open(DATAFILE, 'rb') as f0:
            f0.seek(BYTEOFFSET)
            # TODO: fix this...
            f1 = StringIO(f0.read())
            tr = read(f, format='MSEED')[0]  # f1?
    else:
        tr = read(DATAFILE, format='MSEED')[0]

    return tr.data


def read_s3(DATAFILE, BYTEOFFSET, NUM):
    """
    Read big-endian 24-bit integers.

    Wrapper for Richard Stead's "s3tos4" C routine.

    Returns 32-bit integer NumPy array.

    """
    # open, jump into file
    with open(DATAFILE, 'rb') as f:
        f.seek(BYTEOFFSET, 0)
        # read NUM 3-byte ints
        buf = f.read(NUM * 3)

    # embed 3-byte int buffer into the beginning of a 4-byte string container
    cbuf = C.create_string_buffer(buf, size=NUM * 4)

    # convert int24 buffer to int32 string buffer
    # int convdata(void *buf, int n, char intype, char outtype)
    # retval = libconvert.convdata(cbuf, NUM, C.create_string_buffer('s3', 2),
    #                             C.create_string_buffer('s4', 2))
    retval = libconvert.s3tos4(cbuf, NUM)

    # return as 4-btye ints
    return np.fromstring(cbuf.raw, dtype='>i4'), retval
    # np.frombuffer won't copy the memory, could be useful if fromstring is slow
    # return np.frombuffer(cbuf, dtype='>i4'), retval


def numpy_read(DATAFILE, BYTEOFFSET, NUM, PERMISSION, DTYPE):
    """
    Read NumPy-compatible binary data.

    Modeled after MatSeis function read_file in util/waveread.m.

    """
    f = open(DATAFILE, PERMISSION)
    f.seek(BYTEOFFSET, 0)
    data = np.fromfile(f, dtype=np.dtype(DTYPE), count=NUM)
    f.close()

    return data




#def dbinit(db,prefix):
#    """
#    Command-line arguments are created and parsed, fed to functions.
#
#    """
#    options = get_options(db,prefix)
#    session = get_session(options)
#    files = get_files(options)
#
#    tables = get_or_create_tables(options, session, create=True)
#
#    lastids = ['arid', 'chanid', 'evid', 'orid', 'wfid']
#    last = get_lastids(session, tables['lastid'], lastids, create=True)   
   



from pisces.schema.util import CoreTable
import pisces.schema.kbcore as kb


class Affiliation(kb.Affiliation):
    __tablename__ = 'affiliation'

class Amplitude(kb.Amplitude):
    __tablename__ = 'amplitude'

class Arrival(kb.Arrival):
    __tablename__ = 'arrival'

class Assoc(kb.Assoc):
    __tablename__ = 'assoc'

class Event(kb.Event):
    __tablename__ = 'event'

class Gregion(kb.Gregion):
    __tablename__ = 'gregion'

class Instrument(kb.Instrument):
    __tablename__ = 'instrument'

class Lastid(kb.Lastid):
    __tablename__ = 'lastid'

class Netmag(kb.Netmag):
    __tablename__ = 'netmag'

class Network(kb.Network):
    __tablename__ = 'network'

class Origerr(kb.Origerr):
    __tablename__ = 'origerr'

class Origin(kb.Origin):
    __tablename__ = 'origin'

class Remark(kb.Remark):
    __tablename__ = 'remark'

class Sensor(kb.Sensor):
    __tablename__ = 'sensor'

class Site(kb.Site):
    __tablename__ = 'site'

class Sitechan(kb.Sitechan):
    __tablename__ = 'sitechan'

class Sregion(kb.Sregion):
    __tablename__ = 'sregion'

class Stamag(kb.Stamag):
    __tablename__ = 'stamag'

class Wfdisc(kb.Wfdisc):
    __tablename__ = 'wfdisc'

class Wftag(kb.Wftag):
    __tablename__ = 'wftag'

CORETABLES = {'affiliation': CoreTable('affiliation', kb.Affiliation, Affiliation),
              'arrival': CoreTable('arrival', kb.Arrival, Arrival),
              'assoc': CoreTable('assoc', kb.Assoc, Assoc),
              'event': CoreTable('event', kb.Event, Event),
              'instrument': CoreTable('instrument', kb.Instrument, Instrument),
              'lastid': CoreTable('lastid', kb.Lastid, Lastid),
              'origin': CoreTable('origin', kb.Origin, Origin),
              'site': CoreTable('site', kb.Site, Site),
              'sitechan': CoreTable('sitechan', kb.Sitechan, Sitechan),
              'wfdisc': CoreTable('wfdisc', kb.Wfdisc, Wfdisc)}



from pisces.schema.util import CoreTable
import pisces.schema.antelope as antelope


class Affiliation(antelope.Affiliation):
    __tablename__ = 'affiliation'

class Amplitude(antelope.Amplitude):
    __tablename__ = 'amplitude'

class Arrival(antelope.Arrival):
    __tablename__ = 'arrival'

class Assoc(antelope.Assoc):
    __tablename__ = 'assoc'

class Event(antelope.Event):
    __tablename__ = 'event'

class Gregion(antelope.Gregion):
    __tablename__ = 'gregion'

class Instrument(antelope.Instrument):
    __tablename__ = 'instrument'

class Lastid(antelope.Lastid):
    __tablename__ = 'lastid'

class Netmag(antelope.Netmag):
    __tablename__ = 'netmag'

class Network(antelope.Network):
    __tablename__ = 'network'

class Origerr(antelope.Origerr):
    __tablename__ = 'origerr'

class Origin(antelope.Origin):
    __tablename__ = 'origin'

class Remark(antelope.Remark):
    __tablename__ = 'remark'

class Sensor(antelope.Sensor):
    __tablename__ = 'sensor'

class Site(antelope.Site):
    __tablename__ = 'site'

class Sitechan(antelope.Sitechan):
    __tablename__ = 'sitechan'

class Sregion(antelope.Sregion):
    __tablename__ = 'sregion'

class Stamag(antelope.Stamag):
    __tablename__ = 'stamag'

class Wfdisc(antelope.Wfdisc):
    __tablename__ = 'wfdisc'

class Wftag(antelope.Wftag):
    __tablename__ = 'wftag'


CORETABLES = {'affiliation': CoreTable('affiliation', antelope.Affiliation, Affiliation),
              'arrival': CoreTable('arrival', antelope.Arrival, Arrival),
              'assoc': CoreTable('assoc', antelope.Assoc, Assoc),
              'event': CoreTable('event', antelope.Event, Event),
              'instrument': CoreTable('instrument', antelope.Instrument, Instrument),
              'lastid': CoreTable('lastid', antelope.Lastid, Lastid),
              'origin': CoreTable('origin', antelope.Origin, Origin),
              'site': CoreTable('site', antelope.Site, Site),
              'sitechan': CoreTable('sitechan', antelope.Sitechan, Sitechan),
              'wfdisc': CoreTable('wfdisc', antelope.Wfdisc, Wfdisc)}



import os
import sys
import glob
import argparse
import imp

from sqlalchemy import create_engine
import sqlalchemy.exc as exc
import sqlalchemy.orm.exc as oexc

from obspy import read
from obspy.io.sac.core import _is_sac

import pisces as ps
from pisces.util import get_lastids, url_connect
from pisces.tables.kbcore import CORETABLES
import pisces.io.sac as sac


def get_plugins(plugins):
    """
    Returns a list of imported plugin function objects from a list of plugin
    strings, like path/to/module_file:plugin_function .

    """
    plugin_functions = []
    if plugins:
        for plugin_string in plugins:
            try:
                pth, fn = plugin_string.split(':')
            except ValueError:
                msg = "Must specify plugin like: path/to/modulename:plugin_function"
                raise ValueError(msg)
            pth = pth.split(os.path.sep)
            modname = pth.pop(-1)
            f, pathname, descr = imp.find_module(modname, pth)
            mod = imp.load_module(modname, f, pathname, descr)
            plugin_functions.append(getattr(mod, fn))

    return plugin_functions


def get_files(file_list):
    """
    Return a sequence of SAC file names from either a list of file names
    (trivial) or a text file list (presumable because there are too many files
    to use normal shell expansion).

    """
    if len(file_list) == 1 and not _is_sac(file_list[0]):
        #make a generator of non-blank lines
        try:
            listfile = open(file_list[0], 'r')
            files = (line.strip() for line in listfile if line.strip())
        except IOError:
            msg = "{0} does not exist.".format(file_list[0])
            raise IOError(msg)
    else:
        files = file_list

    return files


# TODO: put in util.py ?
def get_or_create_tables(session, prefix=None, create=True, **tables):
    """
    Load or create canonical ORM KB Core table classes.

    Parameters
    ----------
    session : sqlalchemy.orm.Session
    prefix : str
        Prefix for canonical table names, e.g. 'myaccount.' or 'TA_' (no owner/schema).
        Canonical table names are used if no prefix is provided.
    create : bool
        If True, create tables that don't yet exist.

    Canonical table keyword/tablename string pairs, such as site='other.site',
    can override prefix.  For example, you may want to use a different Lastid
    table, so that ids don't start from 1.

    Returns
    -------
    tables : dict
        A mapping between canonical table names and SQLA declarative classes
        with the correct __tablename__.
        e.g. {'origin': Origin, ...}

    """
    # The Plan:
    # 1. For each core table, build or get the table name
    # 2. If it's a vanilla table name, just use a pre-packaged table class
    # 3. If not, try to autoload it.
    # 4. If it doesn't exist, make it from a prototype and create it in the database.

    # TODO: check options for which tables to produce.

    if not prefix:
        prefix = ''

    tabledict = {}
    for coretable in CORETABLES.values():
        # build the table name
        fulltablename = tables.get(coretable.name, prefix + coretable.name)

        # put table classes into the tables dictionary
        if fulltablename == coretable.name:
            # it's a vanilla table name. just use a pre-packaged table class instead of making one.
            tabledict[coretable.name] = coretable.table
        else:
            tabledict[coretable.name] = ps.make_table(fulltablename, coretable.prototype)

        tabledict[coretable.name].__table__.create(session.bind, checkfirst=True)

    session.commit()

    return tabledict


def dicts2rows(dicts, classes):
    for table, dcts in dicts.items():
        cls = classes[table]
        dicts[table] = [cls(**dct) for dct in dcts]

    return dicts


def make_atomic(last, **rows):
    """
    Unify related table instances/row, including: ids, dir, and dfile
    """
    # last is an AttributeDict of {'keyvalue': lastid instance, ...}
    # rows is a dictionary of {'canonical tablename': [list of instances], ...}
    # of _related_ instances from a single SAC header?
    # TODO: check existance of rows before changing their ids.

    #print rows
    # the order matters here

    # for SAC, only 1
    for event in rows.get('event', []):
        # skips if no 'event' key and if 'event' value is empty []
        # XXX: check for existence first
        event.evid = next(last.evid)

    # for SAC, only 1
    for origin in rows.get('origin', []):
        # XXX: check for existance first
        origin.orid = next(last.orid)
        origin.evid = event.evid

    # for SAC, only 1
    for affil in rows.get('affiliation', []):
        pass

    # for SAC, only 1
    for sitechan in rows.get('sitechan', []):
        # XXX: check for existance first
        sitechan.chanid = next(last.chanid)

    # arrivals correspond to assocs
    for (arrival, assoc) in zip(rows.get('arrival', []), rows.get('assoc', [])):
        arrival.arid = next(last.arid)
        arrival.chanid = sitechan.chanid

        assoc.arid = arrival.arid
        assoc.orid = origin.orid

    # for SAC, only 1
    for wfdisc in rows.get('wfdisc', []):
        wfdisc.wfid = next(last.wfid)


def apply_plugins(plugins, **rows):
    for plugin in plugins:
        rows = plugin(**rows)

    return rows


def sac2db(session, tables, files, plugins=None, absolute_paths=False):
    pass

# TODO: make this main also accept a get_iterable and get_row_dicts functions,
#   so it can be renamed to iter2db and re-used in a sac2db.py and miniseed2db.py
def main(**kwargs):
    """
    Command-line arguments are created and parsed, fed to functions.

    Parameters
    ----------
    session : SQLAlchemy.orm.Session instance
    tables : list
        Canonical names for desired database tables.
    prefix : str
        Target core tables using 'account.prefix' naming.
    absolute_paths : bool
    bbfk : bool
        Pull in deast & dnorth info from user7 & user8 header fields
    file_list : str
        Name of a text file containing full SAC file names.
    files : list
        List of SAC file names.

    """
    print("sac2db: {}".format(kwargs))
    session = kwargs['session']

    if kwargs.get('file_list'):
        files = get_files(kwargs['file_list'])
    else:
        files = kwargs['files']

    tables = get_or_create_tables(session, create=True)

    lastids = ['arid', 'chanid', 'evid', 'orid', 'wfid']
    last = get_lastids(session, tables['lastid'], lastids, create=True)

    # for item in iterable:
    for sacfile in files:
        print(sacfile)

        # row_dicts = get_row_dicts(item)

        tr = read(sacfile, format='SAC', debug_headers=True)[0]

        # rows needs to be a dict of lists, for make_atomic
        # row_dicts = get_row_dicts(tr.stats.sac) # put in the whole trace, to determine byte order?
        dicts = sac.sachdr2tables(tr.stats.sac, tables=tables.keys())
        # row_instances = dicts_to_instances(row_dicts, tables)
        rows = dicts2rows(dicts, tables)

        # manage dir, dfile, datatype
        # XXX: hack.  replace with an updated obspy.io.sac later.
        bo = tr.data.dtype.byteorder
        if bo == '<':
            datatype = 'f4'
        elif bo == '>':
            datatype = 't4'
        elif bo == '=':
            if sys.byteorder == 'little':
                datatype = 'f4'
            else:
                datatype = 't4'

        for wf in rows['wfdisc']:
            wf.datatype = datatype
            wf.dfile = os.path.basename(sacfile)
            if kwargs['absolute_paths']:
                idir = os.path.dirname(os.path.realpath(sacfile))
            else:
                idir = os.path.dirname(sacfile)
                # make sure relative paths are non-empty
                if idir == '':
                   idir = '.'
            wf.dir = idir

        # manage the ids
        make_atomic(last, **rows)

        # plugins = get_plugins(options)

        # rows = apply_plugins(plugins, **rows)

        # add rows to the database
        # XXX: not done very elegantly.  some problem rows are simply skipped.
        for table, instances in rows.items():
            if instances:
                # could be empty []
                try:
                    session.add_all(instances)
                    session.commit()
                except exc.IntegrityError as e:
                    # duplicate or nonexistant primary keys
                    session.rollback()
                    print("rollback {}".format(table))
                except exc.OperationalError as e:
                    # no such table, or database is locked
                    session.rollback()
                    print("rollback {}".format(table))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
""" 
Pisces: A Practical Seismological Database Library in Python


"""
import glob
import setuptools
from setuptools import setup, Extension

# from numpy.distutils.core import setup, Extension

with open('README.md') as readme:
    # https://dustingram.com/articles/2018/03/16/markdown-descriptions-on-pypi
    long_description = readme.read()

doclines = __doc__.split("\n")

setup(name='pisces',
    version='0.3.0',
    description='A Practical Seismological Database Library in Python.',
    long_description=long_description,
    long_description_content_type="text/markdown", # setuptools >= 38.6.0
    author='Jonathan MacCarthy',
    author_email='jkmacc@lanl.gov',
    packages=['pisces','pisces.schema','pisces.io','pisces.tables',
              'pisces.commands'],
    url='https://github.com/jkmacc-LANL/pisces',
    download_url='https://github.com/jkmacc-LANL/pisces/tarball/0.3.0',
    keywords = ['seismology', 'geophysics', 'database'],
    install_requires=['numpy','obspy>=1.0','sqlalchemy>=1.0','Click'],
    ext_package='pisces.io.lib',
    ext_modules=[Extension('libecompression', ['pisces/io/src/e_compression/e_compression.c']),
                 Extension('libconvert', glob.glob('pisces/io/src/convert/*.c'))],
    entry_points = """
        [console_scripts]
        pisces=pisces.commands.main:cli
        """,
    license='LANL-MIT',
    platforms=['Mac OS X', 'Linux/Unix'],
    classifiers=['Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Topic :: Scientific/Engineering',
            'Topic :: Database',
            'Topic :: Scientific/Engineering :: Physics'],
)



# Converted to Python 3.5.2 on 01/25/17
# by: Jeremy Webster

from collections import namedtuple

import sqlalchemy as sa
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
try:
    from sqlalchemy.ext.declarative.api import _declarative_constructor
except ImportError:
    # not >0.8
    from sqlalchemy.ext.declarative import _declarative_constructor

# NO SELF IMPORTS!

# TODO: add a .to_dict() method or a dict-like __getitem__/__setitem__, and
#   remove value iteration, so that copying a row looks like
#   Class(inst.to_dict())
#   or
#   Class(**inst)
#   and this works:
#   for (col, val) in inst:
#       print col, val
#
#   This will affect .from_string (i think) and __init__, and _print_format,
#   and __str__
# TODO: add a decorator like
#   https://blogs.gnome.org/danni/2013/03/07/generating-json-from-sqlalchemy-objects/

# TODO: change 'prototype' and 'table' to 'abstract' and 'concrete'?
CoreTable = namedtuple('CoreTable', ['name', 'prototype', 'table'])


def get_infovals(meta, structure, key):
    """
    Get flattened values from column info dictionary for given metadata and
    desired structure.

    If key is not in column info, None is used.

    Returns
    -------
    colnames: list
        Flattened column names.
    vals: list

    Examples
    --------
    >>> from pisces.schema.kbcore import Base
    >>> fields, defaults = get_infovals(Base.metadata, ['site'], 'default')
    >>> print(defaults)
    ['-', -1, 2286324, -999.0, -999.0, -999.0, '-', '-', '-', -99999, -99999, <built-in method now of type object at 0x100601060>]
    >>> print(fields)
    ['sta', 'ondate', 'offdate', 'lat', 'lon', 'elev', 'staname', 'statype', 'refsta', 'dnorth', 'deast', 'lddate']

    """
    # TODO: make this recursive?
    # TODO: make this more efficient

    # cache all tables
    tabledct = dict([(itable.name, itable) for itable in list(meta.tables.values())])
    # cache all unique column info values in metadata in a dict
    colvaldct = {}
    for t in list(meta.tables.values()):
        for c in t.columns:
            colvaldct[c.name] = c.info.get(key, None)

    structvals = []
    colnames = []
    for _, item in enumerate(structure):
        if item in tabledct:
            itabvals = [c.info.get(key, None) for c in tabledct[item].columns]
            structvals.extend(itabvals)
            colnames.extend([c.name for c in tabledct[item].columns])
        elif item in colvaldct:
            structvals.append(colvaldct[item])
            colnames.append(item)

    return colnames, structvals


def string_formatter(meta, structure):
    """
    Get string substitution formatter for given structure and schema.

    If you're looking for a write_flatfile function, this is it.  It takes
    advantage of native python string formatting and file writing.

    Parameters
    ----------
    meta: sqlalchemy.MetaData
        Columns in tables must have info={'format': ?}
    structure: iterable of str
        Strings are known schema objects (tables, columns).

    Returns
    -------
    fmt: str
        String substitution formatter for provided structure within provided
        schema.

    Examples
    --------
    >>> recs = session.query(Origin, sta.lat, sta.lon).limit(10).all()
    >>> fmt = string_formatter(meta.tables.values(), ['origin', 'lat', 'lon'])
    >>> with open('origin_lat_lon.txt', 'w') as f:
            for rec in recs:
                f.write(fmt.format(rec))

    """
    # TODO: rewrite to use get_infovals
    # XXX: fails with some structures
    # o = session.query(Origin).first()
    # string_formatter(KBBase.metadata, ['lat','lon']).format(o) # fails
    # o = session.query(Origin.lat, Origin.lon).first()
    # string_formatter(KBBase.metadata, ['lat','lon']).format(o) # succeeds
    tabledct = dict([(itable.name, itable) for itable in list(meta.tables.values())])
    colfmtdct = {}
    for t in list(meta.tables.values()):
        for c in t.columns:
            colfmtdct[c.name] = c.info.get('format', None)
    structfmt = []
    for idx, item in enumerate(structure):
        if item in tabledct:
            # e.g. {0.attribute_name:format} {1.attribute_name:format} ...
            itabfmt = ' '.join(["{{{}.{}:{}}}".format(idx, c.name, c.info.get('format', ''))
                                for c in tabledct[item].columns])
            structfmt.append(itabfmt)
        elif item in colfmtdct:
            # e.g. {0:format} {1:format}
            structfmt.append("{{{}:{}}}".format(idx, colfmtdct[item]))
    return ' '.join(structfmt)

# -------- Methods for PiscesMeta --------


def _init(self, *args, **kwargs):
    """
    Create a mapped table instance (a row).

    Constructor accepts all correctly-ordered positional arguments for a table as a
    sequence OR a subset of known keyword arguments.  None values replaced with
    __table__ column info['default'] value.

    """
    # Given to the metaclass to become the __init__ for the class, which means things in it are
    # done to the instances (self).

    # this method is a hot mess:
    # TODO: cache hidden properties line _positions, _types, _defaults, ...
    #    in get_tables?  --> No, put this in PiscesMeta.__init__
    # XXX: fails if no column.info dictionary (schema=Base.metadata wasn't
    #   supplied).
    # XXX: perhaps None was the intended value
    # TODO: change init to accept any kwargs, but ignore unknown ones.

    # XXX: fails for attribute name different from column name (e.g. 'yield_' vs 'yield')
    # use self.__mapper__.columns['yield_'].name to get attr-column mapping
    # TODO: I think it want this snippet to be in PiscesMeta.__new__
    # self._attrname = {c.name: a for a,c in self.__mapper__.c.items()} #{col_name: attr_name}
    # self._strfmt = string_formatter(self.__base__.metadata, [self.__table__.name])
    if args:
        # positional value instantiation
        if kwargs:
            raise ValueError("Either positional or keyword arguments accepted.")
        if len(args) != len(self.__table__.columns):
            raise ValueError("Provide a position argument for each column.")
        for c, ival in zip(self.__table__.columns, args):
            dflt = c.info.get('default', None)
            if ival is None:
                if hasattr(dflt, '__call__'):
                    # handle callables, like datetime.datetime.now
                    setattr(self, self._attrname[c.name], dflt())
                else:
                    setattr(self, self._attrname[c.name], dflt)
            else:
                setattr(self, self._attrname[c.name], ival)
    else:
        # keyword value instantiation
        # use SQLA's keyword constructor, then replace None attribute values with defaults
        _declarative_constructor(self, **kwargs)
        for c, ival in [(col, getattr(self, self._attrname[col.name], None)) for col in self.__table__.columns]:
            dflt = c.info.get('default', None)
            if ival is None:
                if hasattr(dflt, '__call__'):
                    setattr(self, self._attrname[c.name], dflt())
                else:
                    setattr(self, self._attrname[c.name], dflt)


def _str(self):
    """Return a schema-aware flat file row representation."""
    # fmt = string_formatter(self.__table__.metadata, [attrname[c.name] for c in self.__table__.columns])
    # fmt = string_formatter(self.__table__.metadata, [self.__table__.name])
    # return fmt.format(self)

    # XXX: i don't know why this works.  I expected an unpacked `self` to work, because it can
    # iterate itself.  Think it has something to do with the way _format_string is constructed.
    return self._format_string.format(*self)


def _repr(self):
    # TODO: make this use the same mechanics as __getitem__
    """
    This class can be used by ``declarative_base``, to add an automatic
    ``__repr__`` method to *all* subclasses of ``Base``. This ``__repr__``
    will represent values as::

        ClassName(pkey_1=value_1, pkey_2=value_2, ..., pkey_n=value_n)

    where ``pkey_1..pkey_n`` are the primary key columns of the mapped table
    with the corresponding values.

    Notes
    -----
    Stolen and modified from http://foobar.lu/wp/2013/07 .

    """
    # doesn't work for children of DeferredReflection
    items = [(col.name, getattr(self, col.name)) for col in
             self.__mapper__.primary_key]

    return "{0}({1})".format(self.__class__.__name__, ', '.join(['{0}={1!r}'.format(*_) for _ in items]))


def from_string(cls, line, default_on_error=None):
    """
    Construct a mapped table instance from correctly formatted flat file line.

    Works with fixed-length fields, separated by a single whitespace.

    Parameters
    ----------
    line: str
        Flat file line (best to remove newline, but maybe not a problem).
    default_on_error: list, optional
        Supply a list of column names that return default values if they
        produce an error during parsing (e.g. lddate).

    Raises
    ------
    ValueError:  Can't properly parse the line.


    Notes
    -----
    default_on_error is useful for malformed fields, but it will also mask
    other problems with line parsing. It's better to pre-process tables 
    to match the table specifications or catch exceptions and isolate 
    these lines.  

    Examples
    --------
    >>> with open('TA.site','r') as ffsite:
            for line in ffsite:
                isite = Site.from_string(line)
    or
    >>> with open('TA.site','r') as ffsite:
            for line in ffsite:
                isite = Site.from_string(line, default_on_error=['lddate'])

    """
    # TODO: use some sort of cached generator expression at load time for this?
    # TODO: check line length before proceeding.
    # TODO: write this as a separate importable function, and just wrap it here?
    vals = []
    pos = 0
    for col, w, parser in [(c.name, c.info['width'], c.info['parse']) for c in cls.__table__.columns]:
        try:
            val = parser(line[pos:pos + w])
            # print "{} '{}'".format(col, line[pos:pos+w])
        except ValueError as e:
            # TODO: remove this clause, in favor or error handling inside info['parse']
            if default_on_error and col in default_on_error:
                # None are converted to defaults during __init__
                # XXX: no it doesn't.  that'd be nice, though.
                val = None
                # val = c.info['default']
            else:
                msg = ", column {}: '{}', positions [{}:{}]".format(col, line[pos:pos + w], pos, pos + w)
                # XXX: breaks for Python 3
                raise type(e)(str(e) + msg)

                # debuggin
                # print("column: {}, value '{}'".format(c.name, line[pos:pos+w]))
                # raise e

        vals.append(val)
        pos += w + 1

    return cls(*vals)


def _getitem(self, i):
    # integer indexing based on column order in __table__
    # helps implement __iter__ behavior as a side-effect
    values = [getattr(self, self._attrname[c.name], c.info.get('default', None))
              for c in self.__table__.columns]
    return values.__getitem__(i)


def _setitem(self, i, val):
    # integer indexing based on column order in __table__
    # helps implement __iter__ behavior as a side-effect
    colname = list(self.__table__.columns.keys())[i]
    setattr(self, self._attrname[colname], val)


def _len(self):
    # needed for _getitem__, i think
    return len(self.__table__.columns)


def _eq(self, other):
    """ True if primary key values are all equal. """
    return all([getattr(self, self._attrname[c.name]) == getattr(other, other._attrname[c.name])
                for c in self.__table__.primary_key.columns])


def _update_docstring(cls):
    s = '\n'.join(["{} ({}) : {!r}\n    {}".format(c.name, c.key, c.type, c.doc or 'No docstring.')
                   for c in cls.__table__.columns])
    s += "\n\nFORMAT STRING:\n{}\n".format(cls._format_string)
    try:
        s += "\n\nSQL CREATE STATEMENT:\n{}\n".format(sa.schema.CreateTable(cls.__table__))
    except sa.exc.CompileError:
        # One or more columns are of a type that doesn't render to strings.
        pass

    return s


class PiscesMeta(DeclarativeMeta):
    def __new__(cls, clsname, parents, dct):

        # child classes will have methods/data put into "dct"
        dct['__init__'] = _init
        dct['__str__'] = _str
        dct['__repr__'] = _repr
        dct['__getitem__'] = _getitem
        dct['__setitem__'] = _setitem
        dct['__len__'] = _len
        dct['__eq__'] = _eq

        dct['from_string'] = classmethod(from_string)

        dct['_column_info_registry'] = {}  # this is a class-level attribute

        # if __tablename__ looks like 'schema.tablename', split it and move schema to __table_args__
        # see http://www.sqlalchemy.org/trac/ticket/2700 for a possibly
        # better way to do this type of thing.
        try:
            schema, tablename = dct['__tablename__'].split('.')
            dct['__tablename__'] = tablename
            SchemaBase = declarative_base(metadata=sa.MetaData(schema=schema))
            # copy the column registry into the new base, for use with DeferredReflection
            for p in parents:
                if getattr(p, '_column_info_registry', {}):
                    SchemaBase._column_info_registry = p._column_info_registry
            parents = (SchemaBase,) + parents
        except KeyError:
            # no __tablename__, __table_args__
            pass
        except ValueError:
            # not a schema-qualified name
            pass

        return super(PiscesMeta, cls).__new__(cls, clsname, parents, dct)

    def __init__(cls, clsname, parents, dct):

        # called once with a new Base
        # called again for each __abstract__ inheriting class
        # called again for each actual class
        # _not_ called for class instances (rows).  The's dct['__init__']

        # this is SQLA's DeclarativeMeta.__init__
        super(PiscesMeta, cls).__init__(clsname, parents, dct)

        # store Column info dictionary in base._column_info_registry
        for key, val in dct.items():
            if isinstance(val, sa.Column):
                try:
                    cls.__base__.__dict__['_column_info_registry'][key] = val.info
                except KeyError:
                    pass

        # for actual ORM classes, add usefull class attributes
        # "cls._attrname" is a dictionary that gives attribute name for _attrname['column name']
        if hasattr(cls, '__table__'):
            cls._attrname = {c.name: a for a, c in list(cls.__mapper__.c.items())}  # {col_name: attr_name}
            cls._format_string = string_formatter(cls.__base__.metadata, [c.name for c in cls.__table__.columns])
            cls.__doc__ = _update_docstring(cls)


# -------- common parser functions for info['parser'] --------
# these return None upon exception, which is later converted to info['default']

def parse_str(s):
    return str(s).strip() or None


def parse_utf(s):
    """ Decode a utf-8 encoded string. """
    return s.strip().decode('utf-8') or None


def parse_float(s):
    return float(s) or None


def parse_int(s):
    # XXX: int('0') will fail
    return int(s)
